
package jp.gr.java_conf.koto.notavacc.lrg;

import java.util.*;
import jp.gr.java_conf.koto.notavacc.*;
import jp.gr.java_conf.koto.notavacc.util.*;
import jp.gr.java_conf.koto.notavacc.dfa.*;
import jp.gr.java_conf.koto.notavacc.parser.Parser.*;
import jp.gr.java_conf.koto.notavacc.parser.Original.*;
import jp.gr.java_conf.koto.notavacc.parser.Parser.Root;

public final class LRTableGenerator extends Module {
    public LRTableGenerator(Environment env) {
        super(env);
    }
    
    private Root root;
    private DeterministicFiniteAutomaton dfa;
    public LRTable generate(Root root, DeterministicFiniteAutomaton dfa) {
        this.root = root;
        this.dfa = dfa;
        
        logger.verbose(environment.processingFile, "generating extended CFG.");
        Map symbolToInitialState = new LinkedHashMap();
        {
            NondeterministicState finalState = new NondeterministicState();
            Iterator it = root.typeDefinitions().iterator();
            while (it.hasNext()) {
                TypeDefinition def = (TypeDefinition) it.next();
                if (def.abstractKeyword() == null) {
                    Symbol symbol = getNonterminal(def);
                    NondeterministicState initialState = new NondeterministicState();
                    link(initialState, def.expression(), finalState);
                    DeterministicConverter converter = new DeterministicConverter();
                    DeterministicState s = converter.createDeterministicState(Collections.singleton(initialState), finalState);
                    s = buildMinimalForm(s);
                    symbolToInitialState.put(symbol, s);
                }
            }
            it = root.aliasDefinitions().iterator();
            while (it.hasNext()) {
                AliasDefinition def = (AliasDefinition) it.next();
                TreeStackSet argumentLabels = TreeStackSet.EMPTY_SET;
                TreeStackSet labels = TreeStackSet.EMPTY_SET;
                if (root.hasExplicitParameterLabel(def)) {
                    argumentLabels = TreeStackSet.push(argumentLabels, "$label");
                } else {
                    labels = TreeStackSet.push(labels, "$label");
                }
                
                Symbol symbol = getNonterminal(def);
                NondeterministicState initialState = new NondeterministicState();
                link(false, TreeStackSet.EMPTY_SET, argumentLabels, labels, initialState, def.expression(), finalState);
                DeterministicConverter converter = new DeterministicConverter();
                DeterministicState s = converter.createDeterministicState(Collections.singleton(initialState), finalState);
                s = buildMinimalForm(s);
                symbolToInitialState.put(symbol, s);
            }
        }
        if (logger.hasError())
            logger.fatal();
        Set allStates = allStates(symbolToInitialState);
        logger.verbose(environment.processingFile, "generated.  {0} states.", new Integer(allStates.size()));
        //reportStates(symbolToInitialState, allStates);
        
        /*logger.verbose(environment.processingFile, "generating minimal form.");
        symbolToInitialState = buildMinimalForm(symbolToInitialState, allStates);
        allStates = allStates(symbolToInitialState);
        logger.verbose(environment.processingFile, "generated.  {0} states.", new Integer(allStates.size()));
        reportStates(symbolToInitialState, allStates);*/
        
        logger.verbose(environment.processingFile, "generating CFG.");
        symbolToInitialState = buildCFG(symbolToInitialState, allStates);
        allStates = allStates(symbolToInitialState);
        logger.verbose(environment.processingFile, "generated.  {0} states.", new Integer(allStates.size()));
        reportStates(symbolToInitialState, allStates);
        
        Map initialStateToSymbol = new HashMap(symbolToInitialState.size());
        Iterator it25 = symbolToInitialState.entrySet().iterator();
        while (it25.hasNext()) {
            Map.Entry entry = (Map.Entry) it25.next();
            initialStateToSymbol.put(entry.getValue(), entry.getKey());
        }
        assert initialStateToSymbol.size() == symbolToInitialState.size();
        
        
        
        logger.verbose(environment.processingFile, "generating LR states.");
        DeterministicState extendedFinalState = new DeterministicState();
        allStates.add(extendedFinalState);
        LRHelper lrHelper;
        {
            Map symbolToInitialLRState = new LinkedHashMap();
            LRStateBuilder builder = new LRStateBuilder(symbolToInitialState);
            Set definitions = new LinkedHashSet();
            Iterator it = root.typeDefinitions().iterator();
            while (it.hasNext()) {
                TypeDefinition def = (TypeDefinition) it.next();
                if (def.parsableKeyword() != null)
                    definitions.add(def);
            }
            
            Map kernelCoresToLRState = new LinkedHashMap();
            it = definitions.iterator();
            while (it.hasNext()) {
                TypeDefinition def = (TypeDefinition) it.next();
                TypeNonterminal symbol = getNonterminal(def);
                DeterministicState initialState = (DeterministicState) symbolToInitialState.get(symbol);
                
                DeterministicState extendedInitialState = new DeterministicState();
                DeterministicState extendedMediumState = new DeterministicState();
                
                extendedInitialState.symbolToLink.put(symbol, new Link(Collections.EMPTY_SET, symbol, def.identifier()));
                extendedInitialState.symbolToState.put(symbol, extendedMediumState);
                allStates.add(extendedInitialState);
                
                extendedMediumState.symbolToLink.put(Terminal.EOF, new Link(Collections.EMPTY_SET, Terminal.EOF, def.identifier()));
                extendedMediumState.symbolToState.put(Terminal.EOF, extendedFinalState);
                allStates.add(extendedMediumState);
                
                LRState state = builder.createLRState(Collections.singleton(extendedInitialState));
                symbolToInitialLRState.put(symbol, state);
            }
            LRState finalState = builder.createLRState(Collections.singleton(extendedFinalState));
            lrHelper = new LRHelper(symbolToInitialLRState, finalState);
        }
        logger.verbose(environment.processingFile, "generated.  {0} states.", new Integer(lrHelper.allLRStates().size()));
        //reportLRStates(lrHelper);
        
        logger.verbose(environment.processingFile, "tagging.");
        tagLRStates(lrHelper, initialStateToSymbol);
        logger.verbose(environment.processingFile, "tagged.");
        
        logger.verbose(environment.processingFile, "delivering look-a-head.");
        {
            LookaheadDeliverer deliverer = new LookaheadDeliverer(symbolToInitialState, allStates);
            deliverer.deliverLookahead(lrHelper);
        }
        logger.verbose(environment.processingFile, "delivered.");
        
        logger.verbose(environment.processingFile, "generating reductions.");
        {
            ReductionGenerator generator = new ReductionGenerator(symbolToInitialState, initialStateToSymbol, allStates);
            generator.generateReductions(lrHelper);
        }
        logger.verbose(environment.processingFile, "generated.");
        reportLRStates(lrHelper);
        
        logger.verbose(environment.processingFile, "generating LR table.");
        LRTable result = buildLRTable(lrHelper);
        logger.verbose(environment.processingFile, "generated.");
        
        return result;
    }
    
    
    private Map nonterminalToHint = new LinkedHashMap();
    private AliasNonterminal getNonterminal(AliasDefinition definition) {
        AliasNonterminal result = new AliasNonterminal(root.getCanonicalName(definition));
        nonterminalToHint.put(result, definition.identifier());
        return result;
    }
    private Token getHint(AliasNonterminal nonterminal) {
        return (Token) nonterminalToHint.get(nonterminal);
    }
    private TypeNonterminal getNonterminal(TypeDefinition definition) {
        TypeNonterminal result = new TypeNonterminal(root.getCanonicalName(definition));
        nonterminalToHint.put(result, definition.identifier());
        return result;
    }
    private Token getHint(TypeNonterminal nonterminal) {
        return (Token) nonterminalToHint.get(nonterminal);
    }
    private Token getHint(Nonterminal nonterminal) {
        if (nonterminal instanceof TypeNonterminal) {
            return getHint((TypeNonterminal) nonterminal);
        } else {
            return getHint((AliasNonterminal) nonterminal);
        }
    }
    
    
    
    private static class Link {
        public Set labels;
        public Symbol symbol;
        public Token hint;
        public Link(Set labels, Symbol symbol, Token hint) {
            this.labels = labels;
            this.symbol = symbol;
            this.hint = hint;
        }
        
        public int hashCode() {
            return labels.hashCode() ^ symbol.hashCode() ^ hint.hashCode();
        }
        public boolean equals(Object rhs) {
            if (rhs == null)
                return false;
            if (rhs instanceof Link)
                equals((Link) rhs);
            return false;
        }
        public boolean equals(Link rhs) {
            return labels.equals(rhs.labels) && symbol.equals(rhs.symbol) && hint.equals(rhs.hint);
        }
        
        public String toString() {
            StringBuffer result = new StringBuffer();
            Iterator it = labels.iterator();
            while (it.hasNext()) {
                String label = (String) it.next();
                result.append(label);
                result.append(':');
            }
            result.append(symbol);
            return result.toString();
        }
    }
    
    private static class NondeterministicState {
        public Token tag;
        
        public MultiMap linkToStates = new MultiMap();
        public void link(Link link, NondeterministicState state) {
            linkToStates.add(link, state);
        }
        
        public Set epsilonMoveStates = new LinkedHashSet();
        public void epsilonLink(NondeterministicState state) {
            this.epsilonMoveStates.add(state);
        }
    }
    
    private void link(NondeterministicState initialState, Expression expression, NondeterministicState finalState) {
        link(false, TreeStackSet.EMPTY_SET, TreeStackSet.EMPTY_SET, TreeStackSet.EMPTY_SET, initialState, expression, finalState);
    }
    private void link(boolean embed, TreeStackSet openedAlias, TreeStackSet argumentLabels, TreeStackSet labels, NondeterministicState initialState, Expression expression, NondeterministicState finalState) {
        if (expression instanceof SelectiveExpression) {
            SelectiveExpression exp = (SelectiveExpression) expression;
            Iterator it = exp.operands().iterator();
            while (it.hasNext()) {
                Expression operand = (Expression) it.next();
                link(embed, openedAlias, argumentLabels, labels, initialState, operand, finalState);
            }
        } else if (expression instanceof SequentialExpression) {
            SequentialExpression exp = (SequentialExpression) expression;
            if (exp.operands().isEmpty()) {
                initialState.epsilonLink(finalState);
            } else {
                Iterator it = exp.operands().iterator();
                while (it.hasNext()) {
                    Expression operand = (Expression) it.next();
                    NondeterministicState tmp;
                    if (it.hasNext())
                        tmp = new NondeterministicState();
                    else
                        tmp = finalState;
                    link(embed, openedAlias, argumentLabels, labels, initialState, operand, tmp);
                    initialState = tmp;
                }
            }
        } else if (expression instanceof PlusExpression) {
            PlusExpression exp = (PlusExpression) expression;
            NondeterministicState tmp1 = new NondeterministicState();
            NondeterministicState tmp2 = new NondeterministicState();
            initialState.epsilonLink(tmp1);
            link(embed, openedAlias, argumentLabels, labels, tmp1, exp.operand(), tmp2);
            tmp2.epsilonLink(finalState);
            tmp2.epsilonLink(tmp1);
        } else if (expression instanceof RestrictorExpression) {
            RestrictorExpression exp = (RestrictorExpression) expression;
            link(embed, openedAlias, argumentLabels, labels, initialState, exp.operand(), finalState);
        } else if (expression instanceof LabeledExpression) {
            LabeledExpression exp = (LabeledExpression) expression;
            if (exp.label().getImage().equals("$label"))
                labels = TreeStackSet.pushAll(labels, argumentLabels);
            else
                labels = TreeStackSet.push(labels, exp.label().getImage());
            link(embed, openedAlias, argumentLabels, labels, initialState, exp.operand(), finalState);
        } else if (expression instanceof EmbedExpression) {
            EmbedExpression exp = (EmbedExpression) expression;
            link(true, openedAlias, argumentLabels, labels, initialState, exp.operand(), finalState);
        } else if (expression instanceof TagExpression) {
            TagExpression exp = (TagExpression) expression;
            
            NondeterministicState tmp = new NondeterministicState();
            tmp.tag = exp.tag();
            
            Expression operand = exp.operand();
            if (operand instanceof StringExpression) {
                link(false, openedAlias, argumentLabels, labels, initialState, operand, tmp);
            } else {
                link(false, openedAlias, argumentLabels, labels, initialState, (IdentifierExpression) operand, tmp);
            }
            tmp.epsilonLink(finalState);
            // DEPENDED: TAG_AFTER_NON_EPSILON_MOVE
        } else if (expression instanceof StringExpression) {
            StringExpression exp = (StringExpression) expression;
            Terminal terminal = dfa.getTerminal(exp);
            Link link = new Link(labels, terminal, exp.string().token());
            initialState.link(link, finalState);
        } else {
            IdentifierExpression exp = (IdentifierExpression) expression;
            link(embed, openedAlias, argumentLabels, labels, initialState, exp, finalState);
        }
    }
    private void link(boolean embed, TreeStackSet openedAlias, TreeStackSet argumentLabels, TreeStackSet labels, NondeterministicState initialState, IdentifierExpression exp, NondeterministicState finalState) {
        Definition def = root.getDefinition(exp.name());
        Token hint = exp.name().identifier();
        if (def instanceof TypeDefinition) {
            TypeDefinition d = (TypeDefinition) def;
            Link link = new Link(labels, getNonterminal(d), hint);
            initialState.link(link, finalState);
        } else if (def instanceof AliasDefinition) {
            AliasDefinition d = (AliasDefinition) def;
            if (!embed || openedAlias.contains(d)) {
                Link link = new Link(labels, getNonterminal(d), hint);
                initialState.link(link, finalState);
            } else {
                openedAlias = TreeStackSet.push(openedAlias, d);
                if (root.hasExplicitParameterLabel(d)) {
                    argumentLabels = labels;
                    labels = TreeStackSet.EMPTY_SET;
                } else {
                    argumentLabels = TreeStackSet.EMPTY_SET;
                }
                link(false, openedAlias, argumentLabels, labels, initialState, d.expression(), finalState);
            }
        } else {
            TokenReservation d = (TokenReservation) def;
            Link link = new Link(labels, new NamedTerminal(root.getCanonicalName(d)), hint);
            initialState.link(link, finalState);
        }
    }
    
    private static class DeterministicState {
        public Token tag;
        
        public boolean isFinal = false;
        public Map symbolToLink = new LinkedHashMap();
        public Map symbolToState = new LinkedHashMap();
        
        private static int numberOfInstance = 0;
        public final int idNumber = numberOfInstance++;
        public String toString() {
            return "State" + idNumber;
        }
    }
    
    private Set allStates(Map symbolToInitialState) {
        Set result = new LinkedHashSet();
        allStates(result, symbolToInitialState.values());
        return result;
    }
    private void allStates(Set result, Collection states) {
        Iterator it4 = states.iterator();
        while (it4.hasNext()) {
            DeterministicState state = (DeterministicState) it4.next();
            if (result.add(state))
                allStates(result, state.symbolToState.values());
        }
    }
    
    private void reportStates(Map symbolToInitialState, Set allStates) {
        if (logger.verboseLevel() < Logger.LEVEL_VERBOSEST)
            return;
        
        logger.verbosest(environment.processingFile, "----States Report----");
        Iterator it = symbolToInitialState.entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry) it.next();
            Symbol symbol = (Symbol) entry.getKey();
            DeterministicState state = (DeterministicState) entry.getValue();
            logger.verbosest("{0} => x{1}", symbol, state);
        }
        it = allStates.iterator();
        while (it.hasNext()) {
            DeterministicState state = (DeterministicState) it.next();
            logger.verbosest("-- {0} --", state);
            
            if (state.isFinal)
                logger.verbosest("    FINAL STATE");
            if (state.tag != null)
                logger.verbosest("    TAG: " + state.tag);
            Iterator it2 = state.symbolToState.entrySet().iterator();
            while (it2.hasNext()) {
                Map.Entry entry = (Map.Entry) it2.next();
                Symbol symbol = (Symbol) entry.getKey();
                DeterministicState s = (DeterministicState) entry.getValue();
                Link link = (Link) state.symbolToLink.get(symbol);
                logger.verbosest("        {0} ==> x{1}", link, s);
            }
        }
    }
    
    
    private class DeterministicConverter {
        private Map substatesToState = new LinkedHashMap();
        private DeterministicState createDeterministicState(Set substates, NondeterministicState finalState) {
            DeterministicState result = (DeterministicState) substatesToState.get(substates);
            if (result == null) {
                result = new DeterministicState();
                substatesToState.put(substates, result);
                bind(substates, finalState, result);
            }
            return result;
        }
        
        private void bind(Set substates, NondeterministicState finalState, DeterministicState target) {
            Set closure = getEpsilonClosure(substates);
            
            // set tag
            Set tags = new LinkedHashSet();
            Iterator it = substates.iterator();
            while (it.hasNext()) {
                NondeterministicState s = (NondeterministicState) it.next();
                tags.add(s.tag);    // DEPENDS: TAG_AFTER_NON_EPSILON_MOVE
            }
            if (!tags.isEmpty()) {
                if (tags.size() == 1) {
                    target.tag = (Token) tags.iterator().next();
                } else {
                    boolean first = true;
                    it = tags.iterator();
                    while (it.hasNext()) {
                        Token tag = (Token) it.next();
                        if (tag != null) {
                            if (first) {
                                logger.error(tag, "An unified state has different tag(s):");
                                first = false;
                            }
                            logger.error("        {0}", tag);
                        }
                    }
                }
            }
            
            // set final
            target.isFinal = closure.contains(finalState);
            
            // set map
            MultiMap symbolToLinks = new MultiMap();
            MultiMap symbolToSubstates = new MultiMap();
            Iterator it0 = closure.iterator();
            while (it0.hasNext()) {
                NondeterministicState substate = (NondeterministicState) it0.next();
                Iterator it3 = substate.linkToStates.entrySet().iterator();
                while (it3.hasNext()) {
                    Map.Entry entry = (Map.Entry) it3.next();
                    Link link = (Link) entry.getKey();
                    Set states = (Set) entry.getValue();
                    symbolToLinks.add(link.symbol, link);
                    symbolToSubstates.addAll(link.symbol, states);
                }
            }
            
            it = symbolToLinks.entrySet().iterator();
            while (it.hasNext()) {
                Map.Entry entry = (Map.Entry) it.next();
                Symbol symbol = (Symbol) entry.getKey();
                Set links = (Set) entry.getValue();
                Set labelKinds = new LinkedHashSet();
                Iterator it2 = links.iterator();
                while (it2.hasNext()) {
                    Link link = (Link) it2.next();
                    labelKinds.add(link.labels);
                }
                
                if (labelKinds.size() == 1) {
                    target.symbolToLink.put(symbol, (Link) links.iterator().next());
                } else {
                    boolean first = true;
                    it2 = links.iterator();
                    while (it2.hasNext()) {
                        Link link = (Link) it2.next();
                        if (first) {
                            logger.error(link.hint, "An unified symbol has different labeling:");
                            first = false;
                        }
                        logger.error(link.hint, "        {0}", link);
                    }
                }
            }
            
            it = symbolToSubstates.entrySet().iterator();
            while (it.hasNext()) {
                Map.Entry entry = (Map.Entry) it.next();
                Symbol symbol = (Symbol) entry.getKey();
                Set set = (Set) entry.getValue();
                DeterministicState newState = createDeterministicState(set, finalState);
                target.symbolToState.put(symbol, newState);
            }
        }
        
        private Set getEpsilonClosure(Set states) {
            Set result = new LinkedHashSet();
            getEpsilonClosure(result, states);
            return result;
        }
        private void getEpsilonClosure(Set result, Set states) {
            Iterator it1 = states.iterator();
            while (it1.hasNext()) {
                NondeterministicState state = (NondeterministicState) it1.next();
                if (result.add(state))
                    getEpsilonClosure(result, state.epsilonMoveStates);
            }
        }
    }
    
    private DeterministicState buildMinimalForm(DeterministicState initialState) {
        Map symbolToInitialState = Collections.singletonMap(null, initialState);
        Set allStates = allStates(symbolToInitialState);
        symbolToInitialState = buildMinimalForm(symbolToInitialState, allStates);
        return (DeterministicState) symbolToInitialState.get(null);
    }
    private Map buildMinimalForm(Map symbolToInitialState, Set allStates) {
        Set initialStates = new HashSet(symbolToInitialState.values());
        
        // initialize
        List groups = new ArrayList(allStates.size());
        Map stateToGroupID = new HashMap();
        {
            Map tagAndFinalToGroupID = new LinkedHashMap();
            Iterator it = allStates.iterator();
            while (it.hasNext()) {
                DeterministicState state = (DeterministicState) it.next();
                String tag = (state.tag == null) ? null : state.tag.getImage();
                Object key = Arrays.asList(new Object[] { tag, new Boolean(state.isFinal) });
                Integer groupID;
                if (initialStates.contains(state)) {
                    groupID = new Integer(groups.size());
                    groups.add(new LinkedList());
                } else {
                    groupID = (Integer) tagAndFinalToGroupID.get(key);
                    if (groupID == null) {
                        groupID = new Integer(groups.size());
                        groups.add(new LinkedList());
                        tagAndFinalToGroupID.put(key, groupID);
                    }
                }
                List states = (List) groups.get(groupID.intValue());
                states.add(state);
                stateToGroupID.put(state, groupID);
            }
        }
        
        // loop
        int lastMoodifiedIndex = -1;
        int index = 0;
        for (;; index++) {
            if (index >= groups.size()) {
                if (lastMoodifiedIndex == -1)
                    break;
                index -= groups.size();
            }
            
            List group = (List) groups.get(index);
            assert !group.isEmpty();
            if (group.size() <= 1) {
                if (index == lastMoodifiedIndex)
                    break;
                continue;
            }
            
            Map keyToStateList = new LinkedHashMap();
            Iterator it = group.iterator();
            while (it.hasNext()) {
                DeterministicState state = (DeterministicState) it.next();
                Set key = new HashSet();
                Iterator it2 = state.symbolToLink.entrySet().iterator();
                while (it2.hasNext()) {
                    Map.Entry entry = (Map.Entry) it2.next();
                    Symbol symbol = (Symbol) entry.getKey();
                    Link link = (Link) entry.getValue();
                    DeterministicState nextState = (DeterministicState) state.symbolToState.get(symbol);
                    key.add(Arrays.asList(new Object[] { symbol, link.labels, stateToGroupID.get(nextState) }));
                }
                List list = (List) keyToStateList.get(key);
                if (list == null) {
                    list = new LinkedList();
                    keyToStateList.put(key, list);
                }
                list.add(state);
            }
            
            if (keyToStateList.size() <= 1) {
                if (index == lastMoodifiedIndex)
                    break;
                continue;
            }
            
            lastMoodifiedIndex = index;
            it = keyToStateList.values().iterator();
            while (it.hasNext()) {
                List states = (List) it.next();
                if (it.hasNext()) {
                    Integer newID = new Integer(groups.size());
                    groups.add(states);
                    Iterator it2 = states.iterator();
                    while (it2.hasNext()) {
                        DeterministicState state = (DeterministicState) it2.next();
                        stateToGroupID.put(state, newID);
                    }
                } else {
                    groups.set(index, states);
                }
            }
        }
        
        // create result states.
        DeterministicState[] result = new DeterministicState[groups.size()];
        for (int i = 0; i < result.length; i++)
            result[i] = new DeterministicState();
        
        for (int i = 0; i < result.length; i++) {
            List group = (List) groups.get(i);
            DeterministicState hot = result[i];
            
            /*logger.verbosest(environment.processingFile, "x{0} <=", hot);
            Iterator it32 = group.iterator();
            while (it32.hasNext()) {
                DeterministicState old = (DeterministicState) it32.next();
                logger.verbosest(environment.processingFile, "        x{0}", old);
            }*/
            
            DeterministicState old = (DeterministicState) group.get(0);
            hot.tag = old.tag;
            
            hot.isFinal = old.isFinal;
            hot.symbolToLink = old.symbolToLink;
            Iterator it2 = old.symbolToState.entrySet().iterator();
            while (it2.hasNext()) {
                Map.Entry entry = (Map.Entry) it2.next();
                Symbol symbol = (Symbol) entry.getKey();
                DeterministicState nextState = (DeterministicState) entry.getValue();
                Integer nextGroupID = (Integer) stateToGroupID.get(nextState);
                
                DeterministicState hotNext = result[nextGroupID.intValue()];
                hot.symbolToState.put(symbol, hotNext);
            }
        }
        
        // create result map.
        Map resultMap = new LinkedHashMap();
        Iterator it = symbolToInitialState.entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry) it.next();
            Symbol symbol = (Symbol) entry.getKey();
            DeterministicState state = (DeterministicState) entry.getValue();
            Integer groupID = (Integer) stateToGroupID.get(state);
            resultMap.put(symbol, result[groupID.intValue()]);
        }
        return resultMap;
    }
    
    private Map stateToMapPreviousStateToLinks(Set allStates) {
        Map result = new LinkedHashMap();
        Iterator it13 = allStates.iterator();
        while (it13.hasNext()) {
            DeterministicState state = (DeterministicState) it13.next();
            Iterator it16 = state.symbolToLink.values().iterator();
            while (it16.hasNext()) {
                Link link = (Link) it16.next();
                DeterministicState nextState = (DeterministicState) state.symbolToState.get(link.symbol);
                MultiMap map = (MultiMap) result.get(nextState);
                if (map == null) {
                    map = new MultiMap();
                    result.put(nextState, map);
                }
                map.add(state, link);
            }
        }
        return result;
    }
    
    private static final Set LABEL_INHERITOR = Collections.singleton("$label");
    private Map buildCFG(Map symbolToInitialState, Set allStates) {
        Map result = new LinkedHashMap(symbolToInitialState);
        
        Map stateToMapPreviousStateToLinks = stateToMapPreviousStateToLinks(allStates);
        
        Iterator it17 = stateToMapPreviousStateToLinks.entrySet().iterator();
        while (it17.hasNext()) {
            Map.Entry entry = (Map.Entry) it17.next();
            DeterministicState state = (DeterministicState) entry.getKey();
            MultiMap map = (MultiMap) entry.getValue();
            
            int count = 0;
            Iterator it18 = map.entrySet().iterator();
            while (it18.hasNext()) {
                Map.Entry entry2 = (Map.Entry) it18.next();
                Set links = (Set) entry2.getValue();
                count += links.size();
            }
            if (count > 1) {
                AnonymousNonterminal anony = new AnonymousNonterminal();
                boolean used = false;
                
                Iterator it19 = map.entrySet().iterator();
                while (it19.hasNext()) {
                    Map.Entry entry2 = (Map.Entry) it19.next();
                    DeterministicState prevState = (DeterministicState) entry2.getKey();
                    Set links = (Set) entry2.getValue();
                    Iterator it20 = links.iterator();
                    while (it20.hasNext()) {
                        Link link = (Link) it20.next();
                        
                        DeterministicState newState = new DeterministicState();
                        prevState.symbolToState.put(link.symbol, newState);
                        newState.tag = state.tag;
                        if (state.isFinal)
                            newState.isFinal = true;
                        
                        if (!state.symbolToState.isEmpty()) {
                            DeterministicState finalState = new DeterministicState();
                            newState.symbolToLink.put(anony, new Link(LABEL_INHERITOR, anony, link.hint));
                            newState.symbolToState.put(anony, finalState);
                            finalState.isFinal = true;
                            used = true;
                        }
                    }
                }
                
                state.tag = null;
                state.isFinal = false;
                if (used)
                    result.put(anony, state);
            }
        }
        
        return result;
    }
    
    
    private static class LRState implements LRTable.LRState {
        public Map symbolToNextState = new LinkedHashMap();
        public Map symbolToNextState() {
            return symbolToNextState;
        }
        public MultiMap symbolToReductions = null;
        public Map symbolToReductions() {
            return symbolToReductions;
        }
        
        public Token tag;
        public Token tag() {
            return tag;
        }
        
        public Set kernelItems = new LinkedHashSet();
        public Set items = new LinkedHashSet();
        public Map coreToItem = new LinkedHashMap();
        public boolean containsFinal = false;
        
        private static int numberOfInstance = 0;
        public final int idNumber = numberOfInstance++;
        public String toString() {
            return "LRState" + idNumber;
        }
    }
    private static class Item {
        public final LRState owner;
        public final DeterministicState core;
        public final boolean kernel;
        
        public Set lookaheads = null;
        public Map symbolToNextItem = new LinkedHashMap();
        public Set deliveredItems = new LinkedHashSet();
        
        public Item(LRState owner, DeterministicState core, boolean kernel) {
            this.owner = owner;
            this.core = core;
            this.kernel = kernel;
        }
        
        public String toString() {
            String k = kernel ? " (kernel)" : "";
            String f = core.isFinal ? " (FINAL)" : "";
            return core + f + " ; " + lookaheads + k;
        }
    }
    
    public class Reduction implements LRTable.Reduction {
        private Nonterminal reducedSymbol;
        private List labelSetList;
        public Reduction(Nonterminal reducedSymbol, List labelSetList) {
            this.reducedSymbol = reducedSymbol;
            this.labelSetList = labelSetList;
        }
        public Nonterminal reducedSymbol() {
            return reducedSymbol;
        }
        public List labelSetList() {
            return labelSetList;
        }
        
        public int hashCode() {
            return reducedSymbol.hashCode() + labelSetList.hashCode();
        }
        public final boolean equals(Object rhs) {
            if (rhs instanceof Reduction)
                return equals((Reduction) rhs);
            return false;
        }
        public boolean equals(Reduction rhs) {
            return reducedSymbol.equals(rhs.reducedSymbol) && labelSetList.equals(rhs.labelSetList);
        }
        
        public String toString() {
            return reducedSymbol + " : " + labelSetList;
        }
    }
    
    private class LRHelper {
        private Map symbolToInitialLRState;
        private LRState finalState;
        public LRHelper(Map symbolToInitialLRState, LRState finalState) {
            this.symbolToInitialLRState = symbolToInitialLRState;
            this.finalState = finalState;
        }
        
        public Map symbolToInitialLRState() {
            return symbolToInitialLRState;
        }
        public LRState finalState() {
            return finalState;
        }
        
        private Set allLRStates = null;
        public Set allLRStates() {
            if (allLRStates == null) {
                allLRStates = new LinkedHashSet();
                allLRStates.add(finalState);
                allLRStates(allLRStates, symbolToInitialLRState.values());
            }
            return allLRStates;
        }
        private void allLRStates(Set result, Collection states) {
            Iterator it5 = states.iterator();
            while (it5.hasNext()) {
                LRState state = (LRState) it5.next();
                if (result.add(state))
                    allLRStates(result, state.symbolToNextState.values());
            }
        }
        
        private Map initialLRStateToSymbol = null;
        public Map initialLRStateToSymbol() {
            if (initialLRStateToSymbol == null) {
                initialLRStateToSymbol = new LinkedHashMap();
                Iterator it44 = symbolToInitialLRState.entrySet().iterator();
                while (it44.hasNext()) {
                    Map.Entry entry = (Map.Entry) it44.next();
                    initialLRStateToSymbol.put(entry.getValue(), entry.getKey());
                }
                assert initialLRStateToSymbol.size() == symbolToInitialLRState.size();
            }
            return initialLRStateToSymbol;
        }
        
        private MultiMap itemToGeneratorKernels = null;
        public MultiMap itemToGeneratorKernels() {
            if (itemToGeneratorKernels == null) {
                itemToGeneratorKernels = new MultiMap();
                Iterator it12 = allLRStates().iterator();
                while (it12.hasNext()) {
                    LRState state = (LRState) it12.next();
                    Iterator it31 = state.kernelItems.iterator();
                    while (it31.hasNext()) {
                        Item kernel = (Item) it31.next();
                        Set done = new HashSet();
                        LinkedList undone = new LinkedList();
                        undone.add(kernel);
                        while (!undone.isEmpty()) {
                            Item item = (Item) undone.removeFirst();
                            if (done.add(item)) {
                                Iterator it33 = item.deliveredItems.iterator();
                                while (it33.hasNext()) {
                                    Item deliveredItem = (Item) it33.next();
                                    itemToGeneratorKernels.add(deliveredItem, kernel);
                                    undone.add(deliveredItem);
                                }
                            }
                        }
                    }
                }
            }
            return itemToGeneratorKernels;
        }
        public Set getGeneratorKernels(Item item) {
            return itemToGeneratorKernels().getValueSet(item);
        }
        
        private MultiMap itemToPreviousItems = null;
        public MultiMap itemToPreviousItems() {
            if (itemToPreviousItems == null) {
                itemToPreviousItems = new MultiMap();
                Iterator it12 = allLRStates().iterator();
                while (it12.hasNext()) {
                    LRState state = (LRState) it12.next();
                    Iterator it22 = state.items.iterator();
                    while (it22.hasNext()) {
                        Item item = (Item) it22.next();
                        Iterator it23 = item.symbolToNextItem.values().iterator();
                        while (it23.hasNext()) {
                            Item nextItem = (Item) it23.next();
                            itemToPreviousItems.add(nextItem, item);
                        }
                    }
                }
            }
            return itemToPreviousItems;
        }
        public Set getPreviousItems(Item item) {
            return itemToPreviousItems().getValueSet(item);
        }
        
        private Map stateToAscendedStateAndSymbol = null;
        public Map stateToAscendedStateAndSymbol() {
            if (stateToAscendedStateAndSymbol == null) {
                stateToAscendedStateAndSymbol = new LinkedHashMap();
                
                LinkedList undone = new LinkedList();
                Iterator it39 = symbolToInitialLRState.values().iterator();
                while (it39.hasNext()) {
                    LRState state = (LRState) it39.next();
                    stateToAscendedStateAndSymbol.put(state, null);
                    undone.add(state);
                }
                
                while (!undone.isEmpty()) {
                    LRState state = (LRState) undone.removeFirst();
                    Iterator it38 = state.symbolToNextState.entrySet().iterator();
                    while (it38.hasNext()) {
                        Map.Entry entry = (Map.Entry) it38.next();
                        Symbol symbol = (Symbol) entry.getKey();
                        LRState nextState = (LRState) entry.getValue();
                        if (!stateToAscendedStateAndSymbol.containsKey(nextState)) {
                            stateToAscendedStateAndSymbol.put(nextState, new Object[] { state, symbol });
                            undone.add(nextState);
                        }
                    }
                }
            }
            return stateToAscendedStateAndSymbol;
        }
        
    }
    
    /*private Set allLRStates(Map symbolToInitialLRState) {
        Set result = new LinkedHashSet();
        allLRStates(result, symbolToInitialLRState.values());
        return result;
    }
    private void allLRStates(Set result, Collection states) {
        Iterator it5 = states.iterator();
        while (it5.hasNext()) {
            LRState state = (LRState) it5.next();
            if (result.add(state))
                allLRStates(result, state.symbolToNextState.values());
        }
    }*/
    
    private void reportLRStates(LRHelper lrHelper) {
        if (logger.verboseLevel() < Logger.LEVEL_VERBOSEST)
            return;
        
        logger.verbosest(environment.processingFile, "----LRStates Report----");
        
        Iterator it = lrHelper.symbolToInitialLRState().entrySet().iterator();
        while (it.hasNext()) {
            Map.Entry entry = (Map.Entry) it.next();
            Symbol symbol = (Symbol) entry.getKey();
            LRState state = (LRState) entry.getValue();
            logger.verbosest("{0} => x{1}", symbol, state);
        }
        
        it = lrHelper.allLRStates().iterator();
        while (it.hasNext()) {
            LRState state = (LRState) it.next();
            logger.verbosest("-- {0} --", state);
            if (state.tag != null)
                logger.verbosest("        TAG: " + state.tag);
            Iterator it2 = state.symbolToNextState.entrySet().iterator();
            while (it2.hasNext()) {
                Map.Entry entry = (Map.Entry) it2.next();
                Symbol symbol = (Symbol) entry.getKey();
                LRState s = (LRState) entry.getValue();
                logger.verbosest("        {0} ==> x{1}", symbol, s);
            }
            logger.verbosest("    -- items --");
            Iterator it34 = state.items.iterator();
            while (it34.hasNext()) {
                Item item = (Item) it34.next();
                logger.verbosest("        x{0}", item);
                
                if (item.core.isFinal)
                    logger.verbosest("            FINAL STATE");
                Iterator it35 = item.core.symbolToState.entrySet().iterator();
                while (it35.hasNext()) {
                    Map.Entry entry = (Map.Entry) it35.next();
                    Symbol symbol = (Symbol) entry.getKey();
                    DeterministicState s = (DeterministicState) entry.getValue();
                    Link link = (Link) item.core.symbolToLink.get(symbol);
                    logger.verbosest("            {0} ==> x{1}", link, s);
                }
                
            }
            if (state.symbolToReductions != null) {
                logger.verbosest("    -- reductions --");
                Iterator it14 = state.symbolToReductions.entrySet().iterator();
                while (it14.hasNext()) {
                    Map.Entry entry = (Map.Entry) it14.next();
                    Symbol symbol = (Symbol) entry.getKey();
                    Set reductions = (Set) entry.getValue();
                    logger.verbosest("        {0} ==>", symbol);
                    Iterator it15 = reductions.iterator();
                    while (it15.hasNext()) {
                        Reduction reduction = (Reduction) it15.next();
                        logger.verbosest("            {0}", reduction);
                    }
                }
            }
        }
    }
    
    private class LRStateBuilder {
        private Map symbolToInitialState;
        public LRStateBuilder(Map symbolToInitialState) {
            this.symbolToInitialState = symbolToInitialState;
        }
        
        private Map kernelCoresToLRState = new HashMap();
        private LRState createLRState(Set kernelCores) {
            LRState result = (LRState) kernelCoresToLRState.get(kernelCores);
            if (result != null)
                return result;
            
            // create and register.
            result = new LRState();
            kernelCoresToLRState.put(kernelCores, result);
            
            {
                Iterator it = kernelCores.iterator();
                while (it.hasNext()) {
                    DeterministicState kernelCore = (DeterministicState) it.next();
                    Item kernel = new Item(result, kernelCore, true);
                    result.kernelItems.add(kernel);
                    result.coreToItem.put(kernelCore, kernel);
                }
            }
            
            // list the next kernelCores.
            MultiMap symbolToNextKernelCores = new MultiMap();
            LinkedList undone = new LinkedList(result.kernelItems);
            while (!undone.isEmpty()) {
                Item item = (Item) undone.removeFirst();
                result.items.add(item);
                if (item.core.isFinal)
                    result.containsFinal = true;
                Iterator it = item.core.symbolToState.entrySet().iterator();
                while (it.hasNext()) {
                    Map.Entry entry = (Map.Entry) it.next();
                    Symbol symbol = (Symbol) entry.getKey();
                    DeterministicState nextKernelCore = (DeterministicState) entry.getValue();
                    Link link = (Link) item.core.symbolToLink.get(symbol);
                    symbolToNextKernelCores.add(symbol, nextKernelCore);
                    if (symbol instanceof Nonterminal) {
                        DeterministicState core = (DeterministicState) symbolToInitialState.get(symbol);
                        Item deliveredItem = (Item) result.coreToItem.get(core);
                        if (deliveredItem == null) {
                            deliveredItem = new Item(result, core, false);
                            result.coreToItem.put(core, deliveredItem);
                            undone.add(deliveredItem);
                        }
                        item.deliveredItems.add(deliveredItem);
                    }
                }
            }
            Iterator it = symbolToNextKernelCores.entrySet().iterator();
            while (it.hasNext()) {
                Map.Entry entry = (Map.Entry) it.next();
                Symbol symbol = (Symbol) entry.getKey();
                Set nextKernelCores = (Set) entry.getValue();
                LRState nextState = createLRState(nextKernelCores);
                result.symbolToNextState.put(symbol, nextState);
                
                Iterator it35 = result.items.iterator();
                while (it35.hasNext()) {
                    Item item = (Item) it35.next();
                    DeterministicState nextCore = (DeterministicState) item.core.symbolToState.get(symbol);
                    if (nextCore != null) {
                        Item nextItem = (Item) nextState.coreToItem.get(nextCore);
                        item.symbolToNextItem.put(symbol, nextItem);
                    }
                }
            }
            
            return result;
        }
    }
    
    private void tagLRStates(LRHelper lrHelper, Map initialStateToSymbol) {
        Iterator it7 = lrHelper.allLRStates().iterator();
        while (it7.hasNext()) {
            LRState s = (LRState) it7.next();
            Set tags = new LinkedHashSet();
            Iterator it10 = s.kernelItems.iterator();
            while (it10.hasNext()) {
                Item kernelItem = (Item) it10.next();
                tags.add(kernelItem.core.tag);
            }
            s.tag = (Token) tags.iterator().next();
            if (tags.size() > 1) {
                Iterator it = getExamplePath(s, lrHelper, initialStateToSymbol).iterator();
                Nonterminal nonterminal = (Nonterminal) it.next();
                it.next();  // state
                boolean hasNoTagging = tags.remove(null);
                StringBuffer buffer = new StringBuffer();
                while (it.hasNext()) {
                    Symbol symbol = (Symbol) it.next();
                    it.next();  // state
                    buffer.append(symbol);
                    if (it.hasNext())
                        buffer.append(' ');
                }
                Token hint = getHint(nonterminal);
                logger.error("{0} with the input: {1}", hint, buffer);
                logger.error(hint, "    has the different tagging:");
                if (hasNoTagging)
                    logger.error(hint, "        no tagging");
                Iterator it11 = tags.iterator();
                while (it11.hasNext()) {
                    Token tag = (Token) it11.next();
                    logger.error("        {0}", tag);
                }
            }
        }
    }
    
    private class LookaheadDeliverer {
        private Map symbolToInitialState;
        private Set allStates;
        public LookaheadDeliverer(Map symbolToInitialState, Set allStates) {
            this.symbolToInitialState = symbolToInitialState;
            this.allStates = allStates;
            buildEmptyAcceptableSet();
            buildFirstSet();
        }
        
        private Set emptyAcceptable;
        private void buildEmptyAcceptableSet() {
            emptyAcceptable = new HashSet();
            Set unfixed = new LinkedHashSet(allStates);
            boolean modified = true;
            while (!unfixed.isEmpty() && modified) {
                modified = false;
                Iterator outer = unfixed.iterator();
                while (outer.hasNext()) {
                    DeterministicState state = (DeterministicState) outer.next();
                    // We try to know whether the state accepts the empty input.
                    
                    int check = -1;      // -1=unmatch the empty input, 0=unfixed, 1=match the empty input.
                    if (state.isFinal) {
                        check = 1;
                    } else {
                        Iterator inner = state.symbolToState.entrySet().iterator();
                        while (inner.hasNext()) {
                            Map.Entry entry = (Map.Entry) inner.next();
                            Symbol symbol = (Symbol) entry.getKey();
                            if (symbol instanceof Nonterminal) {
                                DeterministicState nextState = (DeterministicState) entry.getValue();
                                DeterministicState s = (DeterministicState) symbolToInitialState.get(symbol);
                                if (unfixed.contains(nextState) || unfixed.contains(s)) {
                                    check = 0;
                                } else {
                                    if (emptyAcceptable.contains(nextState) && emptyAcceptable.contains(s)) {
                                        check = 1;
                                        break;
                                    }
                                }
                            }
                        }
                    }
                    if (check != 0) {
                        outer.remove();
                        modified = true;
                        if (check == 1) {
                            emptyAcceptable.add(state);
                        }
                    }
                }
            }
            
            /*if (logger.verboseLevel() >= Logger.LEVEL_VERBOSEST) {
                logger.verbosest(environment.processingFile, "buildEmptyAcceptableSet");
                Iterator it8 = emptyAcceptable.iterator();
                while (it8.hasNext()) {
                    DeterministicState state = (DeterministicState) it8.next();
                    logger.verbosest("        x{0}", state);
                }
            }*/
        }
        public boolean isEmptyAcceptable(DeterministicState state) {
            return emptyAcceptable.contains(state);
        }
        public boolean isEmptyAcceptable(Nonterminal symbol) {
            DeterministicState state = (DeterministicState) symbolToInitialState.get(symbol);
            return isEmptyAcceptable(state);
        }
        public boolean isEmptyAcceptable(Symbol symbol) {
            if (symbol instanceof Terminal)
                return false;
            return isEmptyAcceptable((Nonterminal) symbol);
        }
        
        private MultiMap stateToFirstSet;
        private void buildFirstSet() {
            stateToFirstSet = new MultiMap();
            {
                Iterator it0 = allStates.iterator();
                while (it0.hasNext()) {
                    DeterministicState state = (DeterministicState) it0.next();
                    stateToFirstSet.put(state, new LinkedHashSet());
                }
            }
            
            Set unfixed = new LinkedHashSet(allStates);
            boolean modified = true;
            while (!unfixed.isEmpty() && modified) {
                modified = false;
                Iterator outer = unfixed.iterator();
                while (outer.hasNext()) {
                    DeterministicState state = (DeterministicState) outer.next();
                    Set first = stateToFirstSet.getValueSet(state);
                    boolean fixed = true;
                    Iterator inner = state.symbolToState.entrySet().iterator();
                    while (inner.hasNext()) {
                        Map.Entry entry = (Map.Entry) inner.next();
                        Symbol s = (Symbol) entry.getKey();
                        if (s instanceof Terminal) {
                            modified |= first.add(s);
                        } else {
                            DeterministicState ds = (DeterministicState) symbolToInitialState.get(s);
                            modified |= first.addAll(stateToFirstSet.getValueSet(ds));
                            if (unfixed.contains(ds))
                                fixed = false;
                            
                            if (isEmptyAcceptable(ds)) {
                                ds = (DeterministicState) entry.getValue();
                                modified |= first.addAll(stateToFirstSet.getValueSet(ds));
                                if (unfixed.contains(ds))
                                    fixed = false;
                            }
                        }
                    }
                    if (fixed) {
                        outer.remove();
                        modified = true;
                    }
                }
            }
            
            /*if (logger.verboseLevel() >= Logger.LEVEL_VERBOSEST) {
                logger.verbosest(environment.processingFile, "buildFirstSet");
                Iterator it8 = stateToFirstSet.entrySet().iterator();
                while (it8.hasNext()) {
                    Map.Entry entry = (Map.Entry) it8.next();
                    DeterministicState state = (DeterministicState) entry.getKey();
                    Set first = (Set) entry.getValue();
                    logger.verbosest("        x{0}  {1}", state, first);
                }
            }*/
        }
        public Set first(DeterministicState state) {
            return stateToFirstSet.getValueSet(state);
        }
        
        public void deliverLookahead(LRHelper lrHelper) {
            Iterator it36 = lrHelper.allLRStates().iterator();
            while (it36.hasNext()) {
                LRState state = (LRState) it36.next();
                Iterator it37 = state.items.iterator();
                while (it37.hasNext()) {
                    Item item = (Item) it37.next();
                    assert item.lookaheads == null;
                    item.lookaheads = new LinkedHashSet();
                }
            }
            
            LinkedList jobs = new LinkedList();
            MultiMap itemToLookaheadDependencies = new MultiMap();
            {
                LinkedList undone = new LinkedList();
                {
                    Iterator it4 = lrHelper.symbolToInitialLRState().values().iterator();
                    while (it4.hasNext()) {
                        LRState state = (LRState) it4.next();
                        undone.addAll(state.kernelItems);
                    }
                }
                Set done = new HashSet();
                while (!undone.isEmpty()) {
                    Item item = (Item) undone.removeFirst();
                    if (done.add(item)) {
                        Iterator it3 = item.symbolToNextItem.entrySet().iterator();
                        while (it3.hasNext()) {
                            Map.Entry entry = (Map.Entry) it3.next();
                            Symbol symbol = (Symbol) entry.getKey();
                            Item nextItem = (Item) entry.getValue();
                            itemToLookaheadDependencies.add(item, nextItem);
                            undone.add(nextItem);
                            
                            if (symbol instanceof Nonterminal) {
                                DeterministicState core = (DeterministicState) symbolToInitialState.get(symbol);
                                Item deliveredItem = (Item) item.owner.coreToItem.get(core);
                                jobs.add(new Object[] { deliveredItem, first(nextItem.core) });
                                undone.add(deliveredItem);
                                if (isEmptyAcceptable(nextItem.core))
                                    itemToLookaheadDependencies.add(item, deliveredItem);
                            }
                        }
                    }
                }
            }
            
            /*Iterator it9 = itemToLookaheadDependencies.entrySet().iterator();
            while (it9.hasNext()) {
                Map.Entry entry = (Map.Entry) it9.next();
                List key = (List) entry.getKey();
                Set values = (Set) entry.getValue();
                System.out.println(key + " -> " + values);
            }*/
            
            Iterator it1 = jobs.iterator();
            while (it1.hasNext()) {
                Object[] job = (Object[]) it1.next();
                Item item = (Item) job[0];
                Set lookaheads = (Set) job[1];
                deliverLookahead(itemToLookaheadDependencies, item, lookaheads);
            }
        }
        private void deliverLookahead(MultiMap itemToLookaheadDependencies, Item item, Set lookaheads) {
            if (item.lookaheads.addAll(lookaheads)) {
                Set dependencies = itemToLookaheadDependencies.getValueSet(item);
                if (dependencies != null) {
                    Iterator it6 = dependencies.iterator();
                    while (it6.hasNext()) {
                        Item dependency = (Item) it6.next();
                        deliverLookahead(itemToLookaheadDependencies, dependency, lookaheads);
                    }
                }
            }
        }
    }
    
    private class ReductionGenerator {
        private Map initialStateToSymbol;
        private Map stateToMapPreviousStateToLinks;
        public ReductionGenerator(Map symbolToInitialState, Map initialStateToSymbol, Set allStates) {
            this.initialStateToSymbol = initialStateToSymbol;
            stateToMapPreviousStateToLinks = stateToMapPreviousStateToLinks(allStates);
        }
        
        
        public void generateReductions(LRHelper lrHelper) {
            Set conflictedStates = new LinkedHashSet();
            Iterator it21 = lrHelper.allLRStates().iterator();
            while (it21.hasNext()) {
                LRState state = (LRState) it21.next();
                assert state.symbolToReductions == null;
                state.symbolToReductions = new MultiMap();
                boolean conflicted = false;
                if (state.containsFinal) {
                    Iterator it5 = state.items.iterator();
                    while (it5.hasNext()) {
                        Item item = (Item) it5.next();
                        if (item.core.isFinal) {
                            Reduction reduction = getReduction(item);
                            Iterator it24 = item.lookaheads.iterator();
                            while (it24.hasNext()) {
                                Symbol lookahead = (Symbol) it24.next();
                                state.symbolToReductions.add(lookahead, reduction);
                            }
                        }
                    }
                    if (!conflicted) {
                        Set shiftableSymbol = state.symbolToNextState.keySet();
                        Iterator it26 = state.symbolToReductions.entrySet().iterator();
                        while (it26.hasNext()) {
                            Map.Entry entry = (Map.Entry) it26.next();
                            Symbol symbol = (Symbol) entry.getKey();
                            Set reductions = (Set) entry.getValue();
                            if (shiftableSymbol.contains(symbol) || reductions.size() > 1)
                                conflicted = true;
                        }
                    }
                }
                if (conflicted)
                    conflictedStates.add(state);
            }
            
            if (!conflictedStates.isEmpty())
                reportConflicts(conflictedStates, lrHelper);
        }
        
        private Reduction getReduction(Item item) {
            LinkedList labelsList = new LinkedList();
            
            DeterministicState state = item.core;
            for (;;) {
                Nonterminal symbol = (Nonterminal) initialStateToSymbol.get(state);
                if (symbol != null)
                    return new Reduction(symbol, labelsList);
                
                MultiMap map = (MultiMap) stateToMapPreviousStateToLinks.get(state);
                state = (DeterministicState) map.keySet().iterator().next();
                Set links = (Set) map.values().iterator().next();
                Link link = (Link) links.iterator().next();
                
                labelsList.addFirst(link.labels);
                
                assert map.size() == 1;
                assert links.size() == 1;
            }
        }
        
        private void reportConflicts(Set conflictedStates, LRHelper lrHelper) {
            if (logger.verboseLevel() < Logger.LEVEL_VERBOSE)
                return;
            
            Set done = new HashSet();
            LinkedList undone = new LinkedList(lrHelper.symbolToInitialLRState().values());
            while (!undone.isEmpty()) {
                LRState state = (LRState) undone.removeFirst();
                if (done.add(state)) {
                    undone.addAll(state.symbolToNextState.values());
                    if (conflictedStates.contains(state)) {
                        reportConflict(state, lrHelper);
                    }
                }
            }
        }
        private void reportConflict(LRState state, LRHelper lrHelper) {
            Set nextSymbols = new LinkedHashSet();
            {
                Set shiftableSymbol = state.symbolToNextState.keySet();
                Iterator it26 = state.symbolToReductions.entrySet().iterator();
                while (it26.hasNext()) {
                    Map.Entry entry = (Map.Entry) it26.next();
                    Symbol symbol = (Symbol) entry.getKey();
                    Set reductions = (Set) entry.getValue();
                    if (shiftableSymbol.contains(symbol) || reductions.size() > 1)
                        nextSymbols.add(symbol);
                }
            }
            
            List path = getExamplePath(state, lrHelper, initialStateToSymbol);
            TreeStack stack = TreeStack.pushAll(TreeStack.EMPTY_STACK, path);
            
            Iterator it = path.iterator();
            Nonterminal nonterminal = (Nonterminal) it.next();
            it.next();  // state
            StringBuffer buffer = new StringBuffer();
            while (it.hasNext()) {
                Symbol symbol = (Symbol) it.next();
                it.next();  // state
                buffer.append(symbol);
                if (it.hasNext())
                    buffer.append(' ');
            }
            
            Token hint = getHint(nonterminal);
            logger.verbose("LALR(1) conflict: {0} with the input:", hint);
            logger.verbose(hint, "        {0}", buffer);
            
            stack = TreeStack.pop(stack);   // last state
            Symbol lastSymbol = (Symbol) stack.top();
            stack = TreeStack.pop(stack);   // last symbol
            
            Iterator it27 = nextSymbols.iterator();
            while (it27.hasNext()) {
                Symbol lookahead = (Symbol) it27.next();
                logger.verbose(hint, "    and the look-a-head {0}, where the reducible noterminals are:", lookahead);
                boolean shiftReduceConflict = state.symbolToNextState.containsKey(lookahead);
                
                reportConflictReduction(stack, state, TreeStack.push(TreeStack.EMPTY_STACK, lastSymbol), lookahead);
                if (shiftReduceConflict) {
                    logger.verbose(hint, "        and shiftable.");
                }
            }
        }
        private void reportConflictReduction(TreeStack stack, LRState state, TreeStack tail, Symbol lookahead) {
            Iterator it28 = state.symbolToReductions.getValueSet(lookahead).iterator();
            while (it28.hasNext()) {
                Reduction reduction = (Reduction) it28.next();
                reportConflictReduction(stack, reduction, tail, lookahead);
            }
        }
        private void reportConflictReduction(TreeStack stack, Reduction reduction, TreeStack tail, Symbol lookahead) {
            Nonterminal reducedSymbol = reduction.reducedSymbol();
            List labelSetList = reduction.labelSetList();
            for (int i = 0; i < labelSetList.size() - 1; i++) {
                stack = TreeStack.pop(stack);   // state
                Symbol symbol = (Symbol) stack.top();
                stack = TreeStack.pop(stack);
                
                tail = TreeStack.push(tail, symbol);
            }
            
            if (reducedSymbol instanceof AnonymousNonterminal) {
                LRState state = (LRState) stack.top();
                state = (LRState) state.symbolToNextState.get(reducedSymbol);
                reportConflictReduction(stack, state, tail, lookahead);
            } else {
                StringBuffer buffer = new StringBuffer();
                Iterator it45 = tail.iterator();
                while (it45.hasNext()) {
                    Symbol symbol = (Symbol) it45.next();
                    buffer.append(symbol);
                    if (it45.hasNext())
                        buffer.append(' ');
                }
                Token hint = getHint(reducedSymbol);
                logger.verbose("        {0} from {1}", hint, buffer);
            }
        }
    }
    
    
    private LRTable buildLRTable(LRHelper lrHelper) {
        return new GeneratedLRTable(lrHelper.symbolToInitialLRState(), lrHelper.finalState());
    }
    
    private static class GeneratedLRTable extends LRTable {
        private final Map symbolToInitialState;
        private final LRState finalState;
        public GeneratedLRTable(Map symbolToInitialState, LRState finalState) {
            this.symbolToInitialState = Collections.unmodifiableMap(symbolToInitialState);
            this.finalState = finalState;
        }
        
        public Map symbolToInitialState() {
            return symbolToInitialState;
        }
        public LRState finalState() {
            return finalState;
        }
    }
    
    
    private static List getExamplePath(LRState state, LRHelper lrHelper, Map initialStateToSymbol) {
        LinkedList result = new LinkedList();
        Map stateToAscendedStateAndSymbol = lrHelper.stateToAscendedStateAndSymbol();
        
        LRState focusedState = state;
        Set focusedKernelItems = focusedState.kernelItems;
        
        boolean beamed = false;
    loop:
        for (;;) {
            //result.addFirst(focusedKernelItems);
            result.addFirst(focusedState);
            
            {
                Symbol symbol = (Symbol) lrHelper.initialLRStateToSymbol().get(focusedState);
                if (symbol != null) {
                    result.addFirst(symbol);
                    break loop;
                }
            }
            
            Object[] ascendedStateAndSymbol = (Object[]) stateToAscendedStateAndSymbol.get(focusedState);
            assert ascendedStateAndSymbol != null;
            LRState ascendedState = (LRState) ascendedStateAndSymbol[0];
            Symbol symbol = (Symbol) ascendedStateAndSymbol[1];
            
            result.addFirst(symbol);
            
            Set ascendedItems = new LinkedHashSet();
            Iterator it40 = focusedKernelItems.iterator();
            while (it40.hasNext()) {
                Item item = (Item) it40.next();
                Iterator it42 = lrHelper.getPreviousItems(item).iterator();
                while (it42.hasNext()) {
                    Item previousItem = (Item) it42.next();
                    if (previousItem.owner.equals(ascendedState) && item.equals(previousItem.symbolToNextItem.get(symbol)))
                        ascendedItems.add(previousItem);
                }
            }
            assert !ascendedItems.isEmpty();
            if (ascendedItems.size() == 1)
                beamed = true;
            
            if (beamed) {
                Iterator it43 = ascendedItems.iterator();
                while (it43.hasNext()) {
                    Item item = (Item) it43.next();
                    Symbol s = (Symbol) initialStateToSymbol.get(item.core);
                    if (s instanceof TypeNonterminal || s instanceof AliasNonterminal) {
                        //result.addFirst(ascendedItems);
                        result.addFirst(ascendedState);
                        result.addFirst(s);
                        break loop;
                    }
                }
            }
            
            Iterator it41 = new ArrayList(ascendedItems).iterator();
            while (it41.hasNext()) {
                Item item = (Item) it41.next();
                Set tmp = lrHelper.getGeneratorKernels(item);
                if (tmp != null)
                    ascendedItems.addAll(tmp);
            }
            ascendedItems.retainAll(ascendedState.kernelItems);
            assert !ascendedItems.isEmpty();
            if (ascendedItems.size() == 1)
                beamed = true;
            
            focusedState = ascendedState;
            focusedKernelItems = ascendedItems;
        }
        
        return result;
    }
}

